/**************************************************************************
*  This file is part of the TCTS project (Tiny Cooperative Task Scheduler)
*
*  Copyright (c) 2014-2016 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Some functionality like TaskCreate, Broadcast, BroadcastAsync
*  comes from the Ethernut (www.ethernut.de) project.
*  Therefore:
*
*  Partial Copyright (c) 2001-2006 by egnite Software GmbH.
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  17.01.2014  mifi  First Version.
*  15.02.2015  mifi  Expanded for Cortex-M.
*  27.09.2015  mifi  Corrected AddTaskToWaitList. TCBFifoAddEnd must used
*                    instead of TCBFifoAddPrio for correct task order.
*  14.11.2015  mifi  Expanded for Cortex-A8.
*  15.11.2015  mifi  Expanded for Nios II.
*  04.01.2016  mifi  Added more timezone functionality.
*                    Corrected "WaitList" bug.
*  10.01.2016  mifi  Rework semaphore handling.
*                    Rework OSTaskCreate, schedule new task if
*                    scheduler is not locked.
*                    Added OSTaskGetPriority and OSTimeDlyUntil.
*                    Added event functionality. 
**************************************************************************/
#define __TCTS_C__

/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/
#include <string.h>
#include <stdint.h>

#include "tcts.h"
#include "tal.h"

/*=======================================================================*/
/*  All extern data                                                      */
/*=======================================================================*/

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

/*
 * Idle and statistic stack size
 */
#if !defined(OS_IDLE_STACK_SIZE)
#define OS_IDLE_STACK_SIZE 512
#endif

#if !defined(OS_STAT_STACK_SIZE)
#define OS_STAT_STACK_SIZE 512
#endif

/*
 * Idle and statistic task priority
 */
#define IDLE_TASK_PRIO     255
#define STAT_TASK_PRIO     2

/*
 * Test if the fifo is empty
 */
#define IsFifoEmpty(_f)    (((_f)->pIn == NULL) ? 1 : 0)

/*=======================================================================*/
/*  Definition of all global Data                                        */
/*=======================================================================*/   

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

/*
 * Ide and Statistic task variables
 */
static OS_STACK(IdleStack, OS_IDLE_STACK_SIZE);
static OS_STACK(StatStack, OS_STAT_STACK_SIZE);

static os_tcb_t TCBIdle;
static os_tcb_t TCBStat;

/*
 * Scheduler lock status
 */
static uint8_t bIsSchedLocked;

/*
 * Time variables
 */
static volatile uint32_t dSystemTick = 0;
static volatile uint32_t dSecTick    = 0;
static volatile uint32_t dUnixtime   = 0;

static int16_t   wTimezoneID           = 0;
static int16_t   wTimezoneOffsetMin    = 0;
static int32_t   dTimezoneOffsetSec    = 0;
static int32_t   dTimezoneDstOffsetSec = 0;

/*
 * Some statistics variables
 */
static uint8_t   bStatEnabled       = 0; 
static uint8_t   bStatCPULoad       = 0;
static uint32_t  dStatTotalTaskTime = 0;
static uint32_t  dStatHirResPeriod  = 0;

/* 
 * The TaskList contains all the tasks of the system. 
 */
static os_tcb_t *pTaskList = NULL;

/* 
 * The ReadyList contains the tasks which are ready to run. 
 */
static os_tcb_fifo_t ReadyList;

/*
 * The WaitList contains the tasks which are waiting
 * with timeout, e.g. OSTimeDly, OSSemaWait or OSEventWait.
 */
static os_tcb_fifo_t WaitList;

/*
 * This is the actual running task.
 */
static os_tcb_t *RunningTask = NULL;

/*
 * This is the task which will run after the context switch.
 */
static os_tcb_t *NewTask = NULL;

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*
 * Check for ARM7 style CPU
 */
#if defined(__ARM_ARCH_4T__) || defined(__ARM_ARCH_7A__)
#include "tcts_arm.c"
#endif

/*
 * Check for Cortex-M style CPU
 */
#if defined(__ARM_ARCH_7EM__) || defined(__ARM_ARCH_7M__)
#include "tcts_cm.c"
#endif

/*
 * Check for Nios II style CPU
 */
#if defined(__NIOS__)
#include "tcts_nios.c"
#endif

/*************************************************************************/
/*  GetStackFreeCount                                                    */
/*                                                                       */
/*  The stack was initialized with 0xCC.                                 */
/*  Unused entries contains still 0xCC.                                  */
/*                                                                       */
/*  In    : pStart, pEnd                                                 */
/*  Out   : none                                                         */
/*  Return: Free entries of the stack                                    */
/*************************************************************************/
static uint32_t GetStackFreeCount (uint8_t *pStart, uint8_t *pEnd)
{
   uint32_t dFreeCount = 0;
   
   while (pStart < pEnd)
   {
      if (0xCC == *pStart)
      {
         dFreeCount++;   
      }
      else
      {
         break;
      }
      pStart++;
   }
   
   return(dFreeCount);
} /* GetStackFreeCount */

/*************************************************************************/
/*  TCBFifoAddPrio                                                       */
/*                                                                       */
/*  Add the task to the FIFO, with respect to the priority.              */ 
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pFifo, pTask                                                 */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void TCBFifoAddPrio (os_tcb_fifo_t *pFifo, os_tcb_t *pTask)
{
   os_tcb_t *pT;
   int         TaskWasAdded = 0;
   
   /* The Idle task must not added to any lists */
   if (&TCBIdle == pTask)
   {
      return;
   }

   /* Check if the fifo is empty */
   if (NULL == pFifo->pIn)
   {
      /* Fifo is empty */
      pTask->pPrev = NULL;    /* Task has no prev */
      pTask->pNext = NULL;    /* and no next */
      
      pFifo->pIn  = pTask;    /* Fifo in and out */      
      pFifo->pOut = pTask;    /* points to the new task */
   }
   else
   {
      pT = pFifo->pIn;
      while (pT != NULL)
      {
         if (pTask->nPrio >= pT->nPrio)
         {
            if (pT->pPrev == NULL)
            {
               /* Insert at the front of the fifo */
               pTask->pPrev = NULL;          /* This is the new first task, therefore no prev. */
               pTask->pNext = pFifo->pIn;    /* Link the new task in front of the fifo. */
               pFifo->pIn->pPrev = pTask;    /* Link the second task (old first) back to the new first task. */
               pFifo->pIn = pTask;           /* Set the fifo->pIn to the new first task. */
            }
            else
            {
               /* Insert in the middle of the fifo */
               pTask->pNext = pT;
               pTask->pPrev = pT->pPrev;
               
               pT->pPrev->pNext = pTask;
               pT->pPrev        = pTask;
            }
            TaskWasAdded = 1;
            break;
         } /* end if (pTask->Prio >= pT->Prio) */
      
         pT = pT->pNext;
      } /* end while (pT != NULL) */
      
      /* Check if task was not added */
      if (0 == TaskWasAdded)
      {
         /* Add Task at the end of the fifo */
         pTask->pNext = NULL;
         pTask->pPrev = pFifo->pOut;
         
         pFifo->pOut->pNext = pTask;
         pFifo->pOut        = pTask;
      }
   }
   
} /* TCBFifoAddPrio */

/*************************************************************************/
/*  TCBFifoRemove                                                        */
/*                                                                       */
/*  Remove a task from the FIFO, output side, and return the task.       */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pFifo                                                        */
/*  Out   : none                                                         */
/*  Return: pTask                                                        */
/*************************************************************************/
static os_tcb_t *TCBFifoRemove (os_tcb_fifo_t *pFifo)
{
   os_tcb_t *pTask = NULL;

   /* Check if the fifo is empty */
   if (pFifo->pOut != NULL)      
   {
      /* Check if only one element is in the fifo */
      if (pFifo->pIn == pFifo->pOut)
      {
         /* Only one element is in the fifo */
         pTask = pFifo->pOut;

         /* Clear the fifo in/out */
         pFifo->pIn  = NULL;
         pFifo->pOut = NULL;
      }
      else
      {
         pTask = pFifo->pOut;          /* Get the last task of the fifo */
         pFifo->pOut = pTask->pPrev;   /* Move the last pointer to the task before the end */
         pFifo->pOut->pNext = NULL;    /* Remove the link from the new last element to the next one */
      }         
   }
   else
   {
      /* Erro, fifo is empty */
      TAL_FAILED();
   }
         
   if (pTask != NULL)
   {
      /* Clear fifo pointer */
      pTask->pPrev = NULL;
      pTask->pNext = NULL;
   }      
         
   return(pTask);
} /* TCBFifoRemove */

/*************************************************************************/
/*  TCBFifoRemoveMiddle                                                  */
/*                                                                       */
/*  Remove a task from the FIFO.                                         */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pFifo, pTask                                                 */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void TCBFifoRemoveMiddle (os_tcb_fifo_t *pFifo, os_tcb_t *pTask)
{
   /* Check if fifo is empty */
   if (NULL == pFifo->pIn)
   {
      /* Error */
      TAL_FAILED();
   }

   /* Check if only one element is in the list */
   if (pFifo->pIn == pFifo->pOut)
   {
      /* Clear the fifo in/out */
      pFifo->pIn  = NULL;
      pFifo->pOut = NULL;
   }
   else
   {
      /* Check if the task is in the middle of the list */
      if      ((pTask->pPrev != NULL) && (pTask->pNext != NULL))
      {
         pTask->pNext->pPrev = pTask->pPrev;
         pTask->pPrev->pNext = pTask->pNext;
      }
      /* Check for the first element */
      else if (pTask->pPrev == NULL)
      {
         pTask->pNext->pPrev = NULL;
         pFifo->pIn = pTask->pNext;
      }
      /* Check for the last element */
      else if (pTask->pNext == NULL)
      {
         pTask->pPrev->pNext = NULL;
         pFifo->pOut = pTask->pPrev;
      }
   }

   pTask->pNext = NULL;
   pTask->pPrev = NULL;   
} /* TCBFifoRemoveMiddle */

/*************************************************************************/
/*  TaskListAdd                                                          */
/*                                                                       */
/*  Add the task to the TaskList.                                        */ 
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pTask                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void TaskListAdd (os_tcb_t *pTask)
{
   /* Check if the list is empty */
   if (NULL == pTaskList)
   {
      /* List is empty */
      pTask->pTaskPrev = NULL;
      pTask->pTaskNext = NULL;

      pTaskList = pTask;
   }
   else
   {
      pTask->pTaskPrev = NULL;
      
      /*
       * Link the new task in front of the fifo.
       */
      pTask->pTaskNext = pTaskList;
      
      /* 
       * Link the old "first task" back to the new task.
       */
      pTaskList->pTaskPrev = pTask;
   
      /*
       * Set the new fifo->pIn.
       */
      pTaskList = pTask;
   }
   
} /* TaskListAdd */

/*************************************************************************/
/*  TaskListRemove                                                       */
/*                                                                       */
/*  Remove a task from the TaskList.                                     */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pTask                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void TaskListRemove (os_tcb_t *pTask)
{
   /* Check if only one element is in the list */
   if ((NULL == pTaskList->pTaskPrev) && (NULL == pTaskList->pTaskNext))
   {
      /* Clear the list */
      pTaskList = NULL;
   }
   else
   {
      /* Check if the task is in the middle of the list */
      if      ((pTask->pTaskPrev != NULL) && (pTask->pTaskNext != NULL))
      {
         pTask->pTaskNext->pTaskPrev = pTask->pTaskPrev;
         pTask->pTaskPrev->pTaskNext = pTask->pTaskNext;
      }
      /* Check for the first element */
      else if (pTask->pTaskPrev == NULL)
      {
         pTask->pTaskNext->pTaskPrev = NULL;
         pTaskList = pTask->pTaskNext;
      }
      /* Check for the last element */
      else if (pTask->pTaskNext == NULL)
      {
         pTask->pTaskPrev->pTaskNext = NULL;
      }
   }
   
} /* TaskListRemove */

/*************************************************************************/
/*  AddTaskToWaitList                                                    */
/*                                                                       */
/*  Add the task to the WaitList fifo.                                   */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pTask                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static __inline__ void AddTaskToWaitList (os_tcb_t *pTask)
{
   os_tcb_fifo_t *pFifo = &WaitList;
    
   pTask->State = OS_TASK_STATE_WAITING;
   
   /*
    * For the correct order of task processing the task must be 
    * added at the end. For example, if there exist three tasks, A, B and C 
    * with the same priority. And the tasks was started in the ABC order the
    * WaitList looks like A->B->C after all tasks was added to the WaitList
    * now.
    *
    * CheckWaitList will move the tasks later from the WaitList to  
    * the ReadyList if the timeout is expired. But this will be
    * done with TCBFifoAddPrio. The ReadyList will look like C->B->A.
    *
    * The task which is started is taken by the scheduler from the 
    * end of the ReadyList. Now the order of task processing is correct.
    */

   /* The Idle task must not added to any lists */
   if (&TCBIdle == pTask)
   {
      return;
   }

   /* Check if the fifo is empty */
   if (NULL == pFifo->pIn)
   {
      /* Fifo is empty */
      pTask->pWaitPrev = NULL;         /* Task has no prev */
      pTask->pWaitNext = NULL;         /* and no next */
      
      pFifo->pIn  = pTask;             /* Fifo in and out */      
      pFifo->pOut = pTask;             /* points to the new task */
   }
   else   
   {
      /* Fifo is not empty */
      pTask->pWaitNext = NULL;         /* This is the new last task, therefore no next. */    
      pTask->pWaitPrev = pFifo->pOut;  /* Link the new task at the end of the fifo. */
      pFifo->pOut->pWaitNext = pTask;  /* Link the old last to the new last. */
      pFifo->pOut = pTask;             /* Set the fifo-pOut to the new last task. */
   }
   
} /* AddTaskToWaitList */

/*************************************************************************/
/*  WaitListRemove                                                       */
/*                                                                       */
/*  Remove a task from the WaitList fifo.                                */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pTask                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void WaitListRemove (os_tcb_t *pTask)
{
   os_tcb_fifo_t *pFifo = &WaitList;
   
   /* Check if fifo is empty */
   if (NULL == pFifo->pIn)
   {
      /* Error */
      TAL_FAILED();
   }
   
   /* Check if only one element is in the list */
   if (pFifo->pIn == pFifo->pOut)
   {
      /* Clear the fifo in/out */
      pFifo->pIn  = NULL;
      pFifo->pOut = NULL;
   }
   else
   {
      /* Check if the task is in the middle of the list */
      if      ((pTask->pWaitPrev != NULL) && (pTask->pWaitNext != NULL))
      {
         pTask->pWaitNext->pWaitPrev = pTask->pWaitPrev;
         pTask->pWaitPrev->pWaitNext = pTask->pWaitNext;
      }
      /* Check for the first element */
      else if (pTask->pWaitPrev == NULL)
      {
         pTask->pWaitNext->pWaitPrev = NULL;
         pFifo->pIn = pTask->pWaitNext;
      }
      /* Check for the last element */
      else if (pTask->pWaitNext == NULL)
      {
         pTask->pWaitPrev->pWaitNext = NULL;
         pFifo->pOut = pTask->pWaitPrev;
      }
   }

   pTask->pWaitPrev = NULL;
   pTask->pWaitNext = NULL;   
} /* WaitListRemove */

/*************************************************************************/
/*  AddTaskToReadyList                                                   */
/*                                                                       */
/*  Add the task to the Ready list                                       */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pTask                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static __inline__ void AddTaskToReadyList (os_tcb_t *pTask)
{
   pTask->State = OS_TASK_STATE_READY;
   TCBFifoAddPrio(&ReadyList, pTask);   
} /* AddTaskToReadyList */

/*************************************************************************/
/*  GetReadyTask                                                         */
/*                                                                       */
/*  Get the task which is ready.                                         */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pTask                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static os_tcb_t *GetReadyTask (void)
{
   os_tcb_t *pTask;
   
   /* In case the ReadyList fifo is empty, return the Idle task */
   if (NULL == ReadyList.pIn)
   {
      pTask = &TCBIdle;
   }
   else
   {
      pTask = TCBFifoRemove(&ReadyList);
   }   
   
   return(pTask);
} /* GetReadyTask */

/*************************************************************************/
/*  StatUpdateTime                                                       */
/*                                                                       */
/*  Update the statistic time information from a task.                   */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pTask                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void StatUpdateTime (os_tcb_t *pTask)
{
   uint16_t wStartTick;
   uint16_t wEndTick;
   uint16_t wStartTimer;
   uint16_t wEndTimer;
   uint16_t wDeltaTick;
   uint16_t wDeltaTimer;
   
   if (bStatEnabled)
   {
      wStartTimer = pTask->dStatStartTime & 0xFFFF;
      wStartTick  = pTask->dStatStartTime >> 16;
      wEndTimer   = pTask->dStatEndTime & 0xFFFF;
      wEndTick    = pTask->dStatEndTime >> 16;
   
      wDeltaTick = wEndTick - wStartTick;

      if (wEndTimer > wStartTimer)
      {
         wDeltaTimer = wEndTimer - wStartTimer;
      }
      else
      {
         wDeltaTimer = (uint16_t)(dStatHirResPeriod - wStartTimer) + wEndTimer;
      }
   
      pTask->dStatTotalTime += (wDeltaTick * dStatHirResPeriod) + wDeltaTimer;
   }  
       
} /* StatUpdateTime */

/*************************************************************************/
/*  CheckWaitList                                                        */
/*                                                                       */
/*  The WaitList is used for handling a timeout or a delay.              */
/*  Check if the task from the WaitList can be moved to the ReadyList.   */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void CheckWaitList (void)
{
   os_tcb_t *pTask = WaitList.pIn;
   os_tcb_t *pNext;

   while (pTask != NULL)
   {
      /* Check if the timeout is expired */
      if (0 == pTask->dTimeoutTicks)
      {
         /* 
          * Save the next pointer of the task.
          * pTask->pNext could not be used later, because it wil be changed
          * by the TCBFifoAdd before we can use the Next for the loop.
          */
         pNext = pTask->pWaitNext;
         
         /* Remove the task from the WaitList */
         WaitListRemove(pTask);
         
         /* Check if the task was waiting for a semaphore */
         if (pTask->pSemaWait)
         {
            /* 
             * This is a timeout of the semaphore, 
             * the task must be removed from the semaphore list too.
             */
            TCBFifoRemoveMiddle(&pTask->pSemaWait->Fifo, pTask);
            pTask->pSemaWait   = NULL;
            pTask->nReturnCode = OS_RC_TIMEOUT;  /* Return code for OSSemaWait */
         }
         
         /* Check if the task was waiting for an event */
         if (pTask->pEventWait )
         {
            /* 
             * This is a timeout of the event, 
             * the task must be removed from the event list too.
             */
            TCBFifoRemoveMiddle(&pTask->pEventWait->Fifo, pTask);
            pTask->pEventWait  = NULL;
            pTask->nReturnCode = OS_RC_TIMEOUT;  /* Return code for OSEventWait */
         }
         
         /* Add the task to the ReadyList */            
         AddTaskToReadyList(pTask);
         
         /* Loop through the list */
         pTask = pNext;
      }
      else
      {
         /* Loop through the list */
         pTask = pTask->pWaitNext;
      }
   } /* end while (pTask != NULL) */
   
} /* CheckWaitList */

/*************************************************************************/
/*  TaskSchedule                                                         */
/*                                                                       */
/*  Must be called with interrupts disabled.                             */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void TaskSchedule (void)
{
   /*
    * Check WaitList
    */
   CheckWaitList(); 
   
   /*
    * Get new task from the ReadyList.
    */
   NewTask = GetReadyTask();
   
   /*
    * If NewTask == RunningTask, we does not need a task switch
    */
   if (NewTask != RunningTask)
   {
      /* Get task end time for statistics */
      RunningTask->dStatEndTime = tal_CPUStatGetHiResCnt();
      StatUpdateTime(RunningTask);
   
      /* Context switch needed */
      ContextSwitch();    

      /* Get task start time for statistics */
      RunningTask->dStatStartTime = tal_CPUStatGetHiResCnt();
   }
   
} /* TaskSchedule */

/*************************************************************************/
/*  StatTask                                                             */
/*                                                                       */
/*  In    : Task parameter                                               */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
static void StatTask (void *pParam)
{   
   os_tcb_t *pTask;
   uint32_t  dIdleTime = 0;
   
   (void)pParam;
   
   while (1)
   {
      if (bStatEnabled)
      {
         /* Get total task time */
         dStatTotalTaskTime = 0;
      
         pTask = pTaskList;
         while (pTask != NULL)
         {
            /* Count total TaskTime */
            dStatTotalTaskTime += pTask->dStatTotalTime;
         
            /* Find Idle time */
            if (IDLE_TASK_PRIO == pTask->nPrio)
            {
               dIdleTime = pTask->dStatTotalTime;
            }
      
            pTask->dStatTotalLastTime = pTask->dStatTotalTime;
            pTask->dStatTotalTime     = 0;
         
            pTask = pTask->pTaskNext;
         }
      
         bStatCPULoad = 100 - (uint8_t)((dIdleTime * 100) / dStatTotalTaskTime);
      }         
      
      OSTimeDly(1000);
   }
   
} /* StatTask */

/*************************************************************************/
/*  IdleTask                                                             */
/*                                                                       */
/*  In    : Task parameter                                               */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
static void IdleTask (void *pParam)
{   
   (void)pParam;
   
   while (1)
   {
      OSTaskYield();
   }
   
} /* IdleTask */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  OSInit                                                               */
/*                                                                       */
/*  Init the "Cooperative Task Scheduler".                               */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSInit (void)
{
   /* Init the Ready and Wait List */
   memset(&ReadyList, 0x00, sizeof(ReadyList));
   memset(&WaitList,  0x00, sizeof(WaitList));
   
   bIsSchedLocked = 1;     /* Used in tcts_cm, _arm and _nios */

   /* Create the Idle task */   
   OSTaskCreate(&TCBIdle, IdleTask, NULL, IDLE_TASK_PRIO, 
                IdleStack, sizeof(IdleStack), "[idle]"); 

   /* Create the Statistic task */   
   OSTaskCreate(&TCBStat, StatTask, NULL, STAT_TASK_PRIO, 
                StatStack, sizeof(StatStack), "[stat]");                
                
   (void)bIsSchedLocked;   /* Prevent lint warning */
} /* OSInit */

/*************************************************************************/
/*  OSOutputRuntimeStackInfo                                             */
/*                                                                       */
/*  Output the runtime stack information.                                */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSOutputRuntimeStackInfo (void)
{
   OutputRuntimeStackInfo();   
} /* OSOutputRuntimeStackInfo */

/*************************************************************************/
/*  OSOutputTaskInfo                                                     */
/*                                                                       */
/*  Output the Stack and Time information from the tasks.                */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSOutputTaskInfo (void)
{
   os_tcb_t *pTCB;
   uint8_t  *pStart;
   uint8_t  *pEnd;
   uint32_t  dSize = 0;
   uint32_t  dFree = 0;
   uint32_t  dTotalTime;
   uint16_t  wCPUTimePercent;
   uint16_t  wTime1 = 0;
   uint16_t  wTime2 = 0;
   
   TAL_PRINTF("*** Task Info ***\n");
   TAL_PRINTF("\n");
   TAL_PRINTF("Task              Prio   Size   Used   Free   CPU %%\n");
   TAL_PRINTF("====================================================\n");

   /* Get start of list */
   pTCB = OSTaskGetList(); 
   
   dTotalTime = OSStatGetTotalTaskTime();

   while (pTCB != NULL)
   {
      if (pTCB->Name[0] != 0)
      {
         TAL_PRINTF("%-16s  ", pTCB->Name);
      }
      else
      {
         TAL_PRINTF("%-16s  ", "----------------");
      }

      pStart = pTCB->pStackStart;
      pEnd   = (uint8_t*)pStart + pTCB->wStackSize;
      dSize  = (uint32_t)pEnd - (uint32_t)pStart;
      dFree  = GetStackFreeCount(pStart, pEnd);
      
      wCPUTimePercent = (uint16_t)(((uint64_t)pTCB->dStatTotalLastTime * 1000) / dTotalTime);
      wTime1 = wCPUTimePercent / 10;
      wTime2 = wCPUTimePercent % 10;
      
      TAL_PRINTF("%4d   %4d   %4d   %4d   %2d.%01d\n", pTCB->nPrio, dSize, (dSize - dFree), dFree, wTime1, wTime2);
      
      pTCB = pTCB->pTaskNext;
   }   
         
   TAL_PRINTF("\n--- CPU load: %d%% ---\n", OSStatGetCPULoad());
         
   (void)dSize;
   (void)dFree;
   (void)wTime1;
   (void)wTime2;
         
   TAL_PRINTF("\n");   
} /* OSOutputTaskInfo */

/*************************************************************************/
/*  OSStatEnable                                                         */
/*                                                                       */
/*  Enable the statistic functionality.                                  */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSStatEnable (void)
{
   bStatEnabled      = 1;
   dStatHirResPeriod = tal_CPUStatGetHiResPeriod();
} /* OSStatEnable */

/*************************************************************************/
/*  OSStatGetCPULoad                                                     */
/*                                                                       */
/*  Return the CPU load in percent.                                      */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
uint8_t OSStatGetCPULoad (void)
{
   return(bStatCPULoad);
} /* OSStatGetCPULoad */

/*************************************************************************/
/*  OSStatGetTotalTaskTime                                               */
/*                                                                       */
/*  Return the total task time, of all tasks, from the last second.      */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
uint32_t OSStatGetTotalTaskTime (void)
{
   return(dStatTotalTaskTime);
} /* OSStatGetTotalTaskTime */

/*************************************************************************/
/*  OSTaskGetPriority                                                    */
/*                                                                       */
/*  Get the priority of the actual running task.                         */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: The task priority                                            */
/*************************************************************************/
int OSTaskGetPriority (void)
{
   return(RunningTask->nPrio);
} /* OSTaskGetPriority */

/*************************************************************************/
/*  OSTaskChangePriority                                                 */
/*                                                                       */
/*  Change the priority of the actual running task.                      */
/*                                                                       */
/*  In    : Prio                                                         */
/*  Out   : none                                                         */
/*  Return: The old priority                                             */
/*************************************************************************/
int OSTaskChangePriority (int nPrio)
{
   int nOldPrio = RunningTask->nPrio;
   
   RunningTask->nPrio = nPrio;
   
   return(nOldPrio);
} /* OSTaskChangePriority */

/*************************************************************************/
/*  OSTaskYield                                                          */
/*                                                                       */
/*  Give CPU time to the next task from the ReadyList.                   */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSTaskYield (void)
{
   EnterCritical();
   
   /*
    * Add the actual running task to the ReadyList.
    */
   AddTaskToReadyList(RunningTask);    
   
   /*
    * Schedule the next task
    */
   TaskSchedule();
   
   ExitCritical();             
} /* OSTaskYield */

/*************************************************************************/
/*  OSTaskExit                                                           */
/*                                                                       */
/*  Terminate the actual running task.                                   */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSTaskExit (void)
{
   EnterCritical();
   
   TaskListRemove(RunningTask);
   RunningTask->State = OS_TASK_STATE_NOT_IN_USE;
   
   /* Schedule to the next task which is ready */
   TaskSchedule();
   
   ExitCritical();      
} /* OSTaskExit */

/*************************************************************************/
/*  OSTaskGetList                                                        */
/*                                                                       */
/*  Return the first element of the task list.                           */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: first element of the task list                               */
/*************************************************************************/
os_tcb_t *OSTaskGetList (void)
{
   return(pTaskList);
} /* OSTaskGetList */

/*************************************************************************/
/*  OSTimerCallback                                                      */
/*                                                                       */
/*  This function will be called from the Systick interrupt.             */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSTimerCallback (void)
{
   os_tcb_t     *pTask = WaitList.pIn;
   static uint32_t dSecCnt = 0;
   
   /* 
    * System ticker 
    */
   dSystemTick++;

   /* 
    * Second ticker
    */
   dSecCnt++;
   if (OS_TICKS_PER_SECOND == dSecCnt)
   {
      dSecCnt = 0;
      dSecTick++;
      dUnixtime++;
   }

   /* 
    * Go through the WaitList and decrease the WaitCount. 
    */
   
   /* Check if WaitList fifo is NOT empty */
   if (WaitList.pIn != NULL)
   {
      pTask = WaitList.pIn;
      while (pTask != NULL)
      {
         if ((pTask->dTimeoutTicks > 0) && (pTask->dTimeoutTicks != OS_WAIT_INFINITE))
         {
            pTask->dTimeoutTicks--;
         }
         
         pTask = pTask->pWaitNext;
      } /* end while (pTask != NULL) */
   } /* if (WaitList.pIn != NULL) */
   
} /* OSTimerCallback */

/*************************************************************************/
/*  OSTimeGet                                                            */
/*                                                                       */
/*  Return the time from the system ticker.                              */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Systick                                                      */
/*************************************************************************/
uint32_t OSTimeGet (void)
{
   return(dSystemTick);
} /* OSTimeGet */

/*************************************************************************/
/*  OSTimeGetSeconds                                                     */
/*                                                                       */
/*  Return the time from the second ticker.                              */
/*                                                                       */
/*  Out   : none                                                         */
/*  Return: dSecTick                                                     */
/*************************************************************************/
uint32_t OSTimeGetSeconds (void)
{
   return(dSecTick);
} /* OSTimeGetSeconds */

/*************************************************************************/
/*  OSTimeDly                                                            */
/*                                                                       */
/*  Causes the current task to wait for a specified interval or, if the  */
/*  specified interval is zero, to give up the CPU for another task      */
/*  (with higher or same priority).                                      */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  In    : TimeoutMs                                                    */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSTimeDly (uint32_t dTimeoutMs)
{
   uint32_t dTicks;
   
   dTicks = OS_MS_2_TICKS(dTimeoutMs);
   
   if (0 == dTicks)
   {
      OSTaskYield();
   }
   else
   {
      EnterCritical();
      RunningTask->dTimeoutTicks = dTicks;
      AddTaskToWaitList(RunningTask);  /* Add the task to the WaitList */
      TaskSchedule();                  /* Schedule the next task */
      ExitCritical();
   }
   
} /* OSTimeDly */

/*************************************************************************/
/*  OSTimeDlyUntil                                                       */
/*                                                                       */
/*  Causes the current task to wait for a specified time.                */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  In    : dTicks                                                       */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSTimeDlyUntil (uint32_t dTicks)
{
   int64_t WaitTime;
   
   /* Calculate delta */
   WaitTime = dTicks - OSTimeGet(); 
   
   if (WaitTime <= 0)
   {
      OSTaskYield();
   }
   else
   {
      EnterCritical();
      RunningTask->dTimeoutTicks = (uint32_t)WaitTime;
      AddTaskToWaitList(RunningTask);  /* Add the task to the WaitList */
      TaskSchedule();                  /* Schedule the next task */
      ExitCritical();
   }
   
} /* OSTimeDlyUntil */

/*************************************************************************/
/*  OSTimeWait                                                           */
/*                                                                       */
/*  Loop for a specified number of milliseconds,                         */
/*  will not release the CPU.                                            */
/*                                                                       */
/*  In    : TimeoutMs                                                    */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSTimeWait (uint32_t dTimeoutMs)
{
   uint32_t dTicks;
   uint32_t dTimeEnd;
   
   dTicks = OS_MS_2_TICKS(dTimeoutMs);
   
   dTimeEnd = dSystemTick + dTicks;
   
   while (dSystemTick <= dTimeEnd)
   {
      __asm__ volatile ("nop");
   }
      
} /* OSTimeWait */

/*************************************************************************/
/*  OSUnixtimeGet                                                        */
/*                                                                       */
/*  Return the unixtime.                                                 */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Unixtime                                                     */
/*************************************************************************/
uint32_t OSUnixtimeGet (void)
{
   return(dUnixtime);
} /* OSUnixtimeGet */

/*************************************************************************/
/*  OSUnixtimeSet                                                        */
/*                                                                       */
/*  Set a new unixtime.                                                  */
/*                                                                       */
/*  In    : dTime                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSUnixtimeSet (uint32_t dTime)
{
   dUnixtime = dTime;
} /* OSUnixtimeSet */

/*************************************************************************/
/*  OSTimezoneIDSet                                                      */
/*                                                                       */
/*  Set the timezone ID.                                                 */
/*                                                                       */
/*  In    : wID                                                          */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSTimezoneIDSet (int16_t wID)
{
   wTimezoneID = wID;
} /* OSTimezoneIDSet */

/*************************************************************************/
/*  OSTimezoneIDGet                                                      */
/*                                                                       */
/*  Return the timezone ID.                                              */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Timezone                                                     */
/*************************************************************************/
int16_t OSTimezoneIDGet (void)
{
   return(wTimezoneID);
} /* OSTimezoneIDGet */

/*************************************************************************/
/*  OSTimezoneMinSet                                                     */
/*                                                                       */
/*  Set the timezone offset to UTC in minutes.                           */
/*                                                                       */
/*  In    : wOffset                                                       */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSTimezoneMinSet (int16_t wOffset)
{
   wTimezoneOffsetMin = wOffset;
   dTimezoneOffsetSec = wOffset * 60;
} /* OSTimezoneMinSet */

/*************************************************************************/
/*  OSTimezoneMinGet                                                     */
/*                                                                       */
/*  Return the timezone offset in minutes to UTC.                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Timezone                                                     */
/*************************************************************************/
int16_t OSTimezoneMinGet (void)
{
   return(wTimezoneOffsetMin);
} /* OSTimezoneMinGet */

/*************************************************************************/
/*  OSTimezoneSecGet                                                     */
/*                                                                       */
/*  Return the timezone offset in seconds to UTC.                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Timezone                                                     */
/*************************************************************************/
int32_t OSTimezoneSecGet (void)
{
   return(dTimezoneOffsetSec);
} /* OSTimezoneSecGet */

/*************************************************************************/
/*  OSTimezoneDstSet                                                     */
/*                                                                       */
/*  Set the daylight saving time offset to UTC in hours.                 */
/*                                                                       */
/*  In    : bOffset                                                      */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSTimezoneDstSet (int8_t bOffset)
{
   dTimezoneDstOffsetSec = bOffset * 60 * 60;
} /* OSTimezoneDstSet */

/*************************************************************************/
/*  OSTimezoneDstSecGet                                                  */
/*                                                                       */
/*  Return the daylight saving time offset in seconds to UTC.            */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Timezone                                                     */
/*************************************************************************/
int32_t OSTimezoneDstSecGet (void)
{
   return(dTimezoneDstOffsetSec);
} /* OSTimezoneDstSecGet */

/*************************************************************************/
/*  OSSemaCreate                                                         */
/*                                                                       */
/*  Create a new semaphore.                                              */
/*                                                                       */
/*  In    : pSema, dCounterStart, dCounterMax                            */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSSemaCreate (os_sema_t *pSema, uint32_t dCounterStart, uint32_t dCounterMax)
{
   memset(pSema, 0x00, sizeof(os_sema_t));

   pSema->dCounter    = dCounterStart;
   pSema->dCounterMax = dCounterMax;   
} /* OSSemaCreate */

/*************************************************************************/
/*  OSSemaDelete                                                         */
/*                                                                       */
/*  Delete a semaphore.                                                  */
/*                                                                       */
/*  In    : pSema                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSSemaDelete (os_sema_t *pSema)
{
   os_tcb_t *pTask;

   EnterCritical();

   /* Check if fifo is not empty  */
   if (0 == IsFifoEmpty(&pSema->Fifo))
   {
      /* Fifo not empty, remove task from the semaphore fifo */
      pTask = TCBFifoRemove(&pSema->Fifo);
      while (pTask != NULL)
      {
         /* Remove semaphore info */
         pTask->pSemaWait = NULL;
      
         /* Removed task from the WaitList too. */
         WaitListRemove(pTask);
         pTask->dTimeoutTicks = 0;

         /* Return code for OSSemaWait */
         pTask->nReturnCode = OS_RC_DELETED;

         /* Make the task which is waiting on the semaphore ready to run */
         AddTaskToReadyList(pTask);

         /* Remove the next task if exist */         
         if (0 == IsFifoEmpty(&pSema->Fifo))
         {
            pTask = TCBFifoRemove(&pSema->Fifo);
         }
         else
         {
            break;
         }   
      }
   }
   
   ExitCritical();   
} /* OSSemaDelete */

/*************************************************************************/
/*  OSSemaSignalFromInt                                                  */
/*                                                                       */
/*  Signal a semaphore from interrupt.                                   */
/*                                                                       */
/*  Must be called with interrupts disabled.                             */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pSema                                                        */
/*  Out   : none                                                         */
/*  Return: The number of tasks woken up / error cause                   */
/*************************************************************************/
int OSSemaSignalFromInt (os_sema_t *pSema)
{
   int        rc = 0;
   os_tcb_t *pTask;
   
   /* Check if fifo is not empty  */
   if (0 == IsFifoEmpty(&pSema->Fifo))
   {
      /* Fifo not empty, remove task from the semaphore fifo */
      pTask = TCBFifoRemove(&pSema->Fifo);

      /* Remove semaphore info */
      pTask->pSemaWait = NULL;
      
      /* Removed the task from the WaitList too. */
      WaitListRemove(pTask);
      pTask->dTimeoutTicks = 0;

      /* Return code for OSSemaWait */
      pTask->nReturnCode = OS_RC_OK;

      /* Make the task which is waiting on the semaphore ready to run */
      AddTaskToReadyList(pTask);
      
      rc = 1;
   }
   else
   {
      /* Fifo was empty, check for max value */
      if (pSema->dCounter < pSema->dCounterMax)
      {
         pSema->dCounter++;
      }
      else
      {
         rc = OS_RC_NO_SPACE;
      }   
   } /* end if (0 == IsFifoEmpty(&pSema->Fifo)) */
   
   return(rc);
} /* OSSemaSignalFromInt */

/*************************************************************************/
/*  OSSemaSignal                                                         */
/*                                                                       */
/*  Signal a semaphore.                                                  */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  In    : pSema                                                        */
/*  Out   : none                                                         */
/*  Return: The number of tasks woken up / error cause                   */
/*************************************************************************/
int OSSemaSignal (os_sema_t *pSema)
{
   int rc;
   
   EnterCritical();
   
   rc = OSSemaSignalFromInt(pSema);
   
   /* Add the task which signal the semaphore to the ReadyList */ 
   AddTaskToReadyList(RunningTask);   

   TaskSchedule(); /* Schedule the next task */
   
   ExitCritical();
   
   return(rc);
} /* OSSemaSignal */

/*************************************************************************/
/*  OSSemaSignalNoSched                                                  */
/*                                                                       */
/*  Signal a semaphore but without scheduling.                           */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  In    : pSema                                                        */
/*  Out   : none                                                         */
/*  Return: The number of tasks woken up / error cause                   */
/*************************************************************************/
int OSSemaSignalNoSched (os_sema_t *pSema)
{
   int rc;
   
   EnterCritical();
   
   rc = OSSemaSignalFromInt(pSema);
   
   ExitCritical();
   
   return(rc);
} /* OSSemaSignalNoSched */

/*************************************************************************/
/*  OSSemaWait                                                           */
/*                                                                       */
/*  Blocks the task while waiting for the semaphore, with timeout.       */
/*                                                                       */
/*  In case the timeout value (dTimeoutMs) is 0, OSSemaWait will use     */
/*  polling to acquire the resource. If the resource counter (dCounter)  */
/*  is 0, the error cause is OS_RC_TIMEOUT.                              */
/*                                                                       */
/*  With a timeout value (dTimeoutMs) of OS_WAIT_INFINITE the waiting    */
/*  time is not ended until the semaphore is signaled.                   */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  In    : pSema, dTimeoutMs                                            */
/*  Out   : none                                                         */
/*  Return: OS_RC_OK / error casue                                       */
/*************************************************************************/
int OSSemaWait (os_sema_t *pSema, uint32_t dTimeoutMs)
{
   int      rc = OS_RC_OK;
   uint32_t dTicks = OS_MS_2_TICKS(dTimeoutMs);
   
   EnterCritical();
   
   /* Check if semaphore is available */
   if (pSema->dCounter > 0)
   {
      pSema->dCounter--;
   }   
   else
   {
      if (dTicks != 0)
      {
         /* Not available, prepare the task for waiting */
         RunningTask->dTimeoutTicks = dTicks;
         RunningTask->pSemaWait     = pSema;
         RunningTask->nReturnCode   = OS_RC_ERROR;
         RunningTask->State         = OS_TASK_STATE_WAITING;

         /* Add the task to the fifo of the semaphore */
         TCBFifoAddPrio(&pSema->Fifo, RunningTask);
      
         /*  Add the task to the WaitList too */
         AddTaskToWaitList(RunningTask);      

         TaskSchedule(); /* Schedule the next task */
      
         /* This is the return code of the function */
         rc = RunningTask->nReturnCode;
      }
      else
      {
         /* Polling, no resource available */
         rc = OS_RC_TIMEOUT; 
      }         
   }
   
   ExitCritical();
   
   return(rc);
} /* OSSemaWait */

/*************************************************************************/
/*  OSSemaBroadcastAsync                            (needed for NutNET)  */
/*                                                                       */
/*  Wake up all tasks waiting on this semaphore. But even if the         */
/*  priority of any woken task is higher than the current tasks's        */
/*  priority, the current one continues running.                         */
/*                                                                       */
/*  Note: No scheduling here, must not used from inside an interrupt.    */
/*                                                                       */
/*  In    : pSema                                                        */
/*  Out   : none                                                         */
/*  Return: The number of tasks woken up / error cause                   */
/*************************************************************************/
int OSSemaBroadcastAsync (os_sema_t *pSema)
{
   int         rc = 0;
   os_tcb_t *pTask;
   
   EnterCritical();

   /* Check if fifo is not empty  */
   if (0 == IsFifoEmpty(&pSema->Fifo))
   {
      /* Fifo not empty, remove task from the semaphore fifo */
      pTask = TCBFifoRemove(&pSema->Fifo);
      while (pTask != NULL) 
      {
         /* Remove semaphore info */
         pTask->pSemaWait = NULL;
      
         /* Removed the task from the WaitList too. */
         WaitListRemove(pTask);
         pTask->dTimeoutTicks = 0;

         /* Return code for OSSemaWait */
         pTask->nReturnCode = OS_RC_OK;

         /* Make the task which is waiting on the semaphore ready to run */
         AddTaskToReadyList(pTask);
         
         rc++;

         /* Get the next task from the fifo */         
         pTask = TCBFifoRemove(&pSema->Fifo);
      }          
   }
   else
   {
      /* Fifo was empty, check for max value */
      if (pSema->dCounter < pSema->dCounterMax)
      {
         pSema->dCounter++;
      }
      else
      {
         rc = OS_RC_NO_SPACE;
      }   
   } /* end if (0 == IsFifoEmpty(&pSema->Fifo)) */
   
   ExitCritical();
   
   return(rc);
} /* OSSemaBroadcastAsync */

/*************************************************************************/
/*  OSSemaBroadcast                                 (needed for NutNET)  */
/*                                                                       */
/*  Wake up all tasks waiting on this semaphore. If the priority of any  */
/*  waiting tasks is higher or equal than the current task's priority,   */
/*  then the current task is stopped and CPU control is passed to the    */
/*  woken up task with the highest priority.                             */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  Out   : none                                                         */
/*  Return: The number of tasks woken up / error cause                   */
/*************************************************************************/
int OSSemaBroadcast (os_sema_t *pSema)
{
   int rc;
   
   rc = OSSemaBroadcastAsync(pSema);
   
   /*
    * If any tasks with higher or equal priority is
    * ready to run, switch the context.
    */
   OSTaskYield();
   
   return(rc);
} /* OSSemaBroadcast */  

/*************************************************************************/
/*  OSEventCreate                                                        */
/*                                                                       */
/*  Create a new event.                                                  */
/*                                                                       */
/*  In    : pEvent                                                       */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSEventCreate (os_event_t *pEvent)
{
   memset(pEvent, 0x00, sizeof(os_event_t));

   pEvent->dPattern = 0;
} /* OSEventCreate */

/*************************************************************************/
/*  OSEventDelete                                                        */
/*                                                                       */
/*  Delete an event.                                                     */
/*                                                                       */
/*  In    : pEvent                                                       */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSEventDelete (os_event_t *pEvent)
{
   os_tcb_t *pTask;

   EnterCritical();

   /* Check if fifo is not empty  */
   if (0 == IsFifoEmpty(&pEvent->Fifo))
   {
      /* Fifo not empty, remove task from the event fifo */
      pTask = TCBFifoRemove(&pEvent->Fifo);
      while (pTask != NULL)
      {
         /* Remove event info */
         pTask->pEventWait = NULL;
         pTask->dEventWaitPattern = 0;
      
         /* Removed the task from the WaitList too. */
         WaitListRemove(pTask);
         pTask->dTimeoutTicks = 0;

         /* Return code for OSEventWait */
         pTask->nReturnCode = OS_RC_DELETED;

         /* Make the task which is waiting on the event ready to run */
         AddTaskToReadyList(pTask);

         /* Remove the next task if exist */         
         if (0 == IsFifoEmpty(&pEvent->Fifo))
         {
            pTask = TCBFifoRemove(&pEvent->Fifo);
         }
         else
         {
            break;
         }   
      }
   }
   
   ExitCritical();
} /* OSEventDelete */

/*************************************************************************/
/*  OSEventSetFromInt                                                    */
/*                                                                       */
/*  Signal an event from interrupt.                                      */
/*                                                                       */
/*  Must be called with interrupts disabled.                             */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pEvent, dPattern                                             */
/*  Out   : none                                                         */
/*  Return: The number of tasks woken up / error cause                   */
/*************************************************************************/
int OSEventSetFromInt (os_event_t *pEvent, uint32_t dPattern)
{
   int        rc = 0;
   int        GoReady;
   os_tcb_t *pTask;
   os_tcb_t *pPrev;

   pEvent->dPattern |= dPattern;
   
   /* Check if fifo is NOT empty  */
   if (0 == IsFifoEmpty(&pEvent->Fifo))
   {
      /* Fifo is not empty */
      pTask = pEvent->Fifo.pOut;

      while (pTask != NULL)
      {
         /* 
          * Save the prev pointer of the task.
          * pTask->pPrev could not be used later, because it wil be changed
          * by the TCBFifoRemoveMiddle before we can use the Prev for the loop.
          */
         pPrev = pTask->pPrev;
      
         /*
          * Check if the task must wait or not
          */
         if (OS_EVENT_MODE_OR == pTask->EventWaitMode)
         {
            GoReady = ((pEvent->dPattern & pTask->dEventWaitPattern) != 0) ? 1 : 0;   
         }
         else
         {
            GoReady = ((pEvent->dPattern & pTask->dEventWaitPattern) == pTask->dEventWaitPattern) ? 1 : 0; 
         }
   
         if (GoReady)
         {
            pTask->dEventWaitPattern = pEvent->dPattern;
            pTask->nReturnCode       = OS_RC_OK;
         
            TCBFifoRemoveMiddle(&pEvent->Fifo, pTask);
         
            /* Remove event info */
            pTask->pEventWait = NULL;
                  
            /* Removed the task from the WaitList too. */
            WaitListRemove(pTask);
            pTask->dTimeoutTicks = 0;
         
            AddTaskToReadyList(pTask);
         
            rc++;
         }
   
         pTask = pPrev;
      }
      
      if (rc != 0)
      {
         /* Event occurred, the pattern can be cleared */
         pEvent->dPattern = 0;
      }
   }
   
   return(rc);
} /* OSEventSetFromInt */

/*************************************************************************/
/*  OSEventSet                                                           */
/*                                                                       */
/*  Signal an event.                                                     */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  In    : pEvent, dPattern                                             */
/*  Out   : none                                                         */
/*  Return: The number of tasks woken up / error cause                   */
/*************************************************************************/
int OSEventSet (os_event_t *pEvent, uint32_t dPattern)
{
   int rc;

   EnterCritical();
   
   rc = OSEventSetFromInt(pEvent, dPattern);
   
   /* Add the task which signal the event to the ReadyList */ 
   AddTaskToReadyList(RunningTask);   

   TaskSchedule(); /* Schedule the next task */
   
   ExitCritical();
   
   return(rc);
} /* OSEventSet */

/*************************************************************************/
/*  OSEventSetNoSched                                                    */
/*                                                                       */
/*  Signal an event but without scheduling.                              */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  In    : pEvent, dPattern                                             */
/*  Out   : none                                                         */
/*  Return: The number of tasks woken up / error cause                   */
/*************************************************************************/
int OSEventSetNoSched (os_event_t *pEvent, uint32_t dPattern)
{
   int rc;

   EnterCritical();
   
   rc = OSEventSetFromInt(pEvent, dPattern);
      
   ExitCritical();
   
   return(rc);
} /* OSEventSetNoSched */

/*************************************************************************/
/*  OSEventWait                                                          */
/*                                                                       */
/*  Blocks the task while waiting for the event, with timeout.           */
/*                                                                       */
/*  With a timeout value (dTimeoutMs) of OS_WAIT_INFINITE the waiting    */
/*  time is not ended until the event is signaled.                       */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  In    : pEvent, dWaitPattern, Mode, pPattern, dTimeoutMs             */
/*  Out   : none                                                         */
/*  Return: OS_EVENT_RC_OK / error casue                                 */
/*************************************************************************/
int OSEventWait (os_event_t *pEvent, uint32_t dWaitPattern, 
                 os_event_mode_t Mode, uint32_t *pPattern, uint32_t dTimeoutMs)
{
   int       rc      = OS_RC_ERROR;
   int       GoReady = 0;
   uint32_t dTicks   = OS_MS_2_TICKS(dTimeoutMs);   
   
   EnterCritical();
   
   *pPattern = 0;
   
   /*
    * Check if the task must wait or not
    */
   if (OS_EVENT_MODE_OR == Mode)
   {
      GoReady = ((pEvent->dPattern & dWaitPattern) != 0) ? 1 : 0;   
   }
   else
   {
      GoReady = ((pEvent->dPattern & dWaitPattern) == dWaitPattern) ? 1 : 0; 
   }
   
   if (GoReady)
   {
      *pPattern = pEvent->dPattern;
      pEvent->dPattern = 0;
      
      rc = OS_RC_OK;
   }
   else
   {
      if (dTicks != 0)
      {
         /* Prepare the task for waiting */
         RunningTask->dTimeoutTicks     = dTicks;
         RunningTask->pEventWait        = pEvent;
         RunningTask->dEventWaitPattern = dWaitPattern;
         RunningTask->EventWaitMode     = Mode;
         RunningTask->nReturnCode       = OS_RC_ERROR;
         RunningTask->State             = OS_TASK_STATE_WAITING;
      
         /* Add the task to the fifo of the event */
         TCBFifoAddPrio(&pEvent->Fifo, RunningTask);
      
         /*  Add the task to the WaitList too */
         AddTaskToWaitList(RunningTask);      

         TaskSchedule(); /* Schedule the next task */

         /* This is the return code of the function */
         rc = RunningTask->nReturnCode;
         if (OS_RC_OK == rc)
         {
            *pPattern = RunningTask->dEventWaitPattern;
         }
      }
      else
      {
         /* Polling, no event available */
         rc = OS_RC_TIMEOUT; 
      }         
   }
      
   ExitCritical();
   
   return(rc);
} /* OSEventWait */

/*** EOF ***/
