/**************************************************************************
*  Copyright (c) 2013-2015 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  19.04.2013  mifi  First Version
**************************************************************************/
#define __BMP085_C__

/*
 * All the information are based on the following document:
 * 
 * Data sheet, BMP085, Digital pressure sensor
 * Version 1.3, 22 August 2011, BST-BMP085-DS000-06
 */
 
/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/
#include <stdint.h>
#include "tal.h"
#include "bmp085.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

/* Use PB10 for Clock */
#define CLK_LOW()    { GPIOB->BSRRH = (1<<10); }
#define CLK_HIGH()   { GPIOB->BSRRL = (1<<10); }

/* Use PB11 for Data */
#define DATA_LOW()   { GPIOB->BSRRH = (1<<11); }
#define DATA_HIGH()  { GPIOB->BSRRL = (1<<11); }
#define DATA_GET()   ((GPIOB->IDR & (1<<11)) ? 1 : 0)

/* Use PB05 for EOC */
#define EOC_GET()    ((GPIOB->IDR & (1<<5)) ? 1 : 0)

#define WAIT_DATA()  { OSTimeDly(2); }
#define WAIT()       { OSTimeDly(2); }

#define START()      {                 \
                        DATA_HIGH();   \
                        WAIT_DATA();   \
                        CLK_HIGH();    \
                        WAIT_DATA();   \
                        DATA_LOW();    \
                        WAIT_DATA();   \
                        CLK_LOW();     \
                        WAIT_DATA();   \
                     }
                       
#define STOP()       {                 \
                        CLK_HIGH();    \
                        WAIT_DATA();   \
                        DATA_HIGH();   \
                        WAIT_DATA();   \
                     }
                       
/*
 * Some defines for the communication with the bmp085
 */
#define ACK             1
#define NO_ACK          0

#define MAX_CONV_TEMP_TIMEOUT    40000
#define MAX_CONV_PRES_TIMEOUT    40000

#define BMP_ADDR_READ   0xEF
#define BMP_ADDR_WRITE  0xEE  


/*=======================================================================*/
/*  Prototypes                                                           */
/*=======================================================================*/

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

/*
 * Calibration data
 */
static int16_t  AC1;
static int16_t  AC2;
static int16_t  AC3;
static uint16_t AC4;
static uint16_t AC5;
static uint16_t AC6;
static int16_t  B1;
static int16_t  B2;
static int16_t  MB;
static int16_t  MC;
static int16_t  MD;

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*************************************************************************/
/*  WriteByte                                                            */
/*                                                                       */
/*  Writes a byte on the bus and checks the acknowledge.                 */
/*                                                                       */
/*  Note: Tested with a STM32F103 where the code was running in flash    */
/*        and the CPU clock was 72MHz.                                   */
/*                                                                       */
/*  In    : Value                                                        */
/*  Out   : none                                                         */
/*  Return: 0 = no error / 1 = error                                     */
/*************************************************************************/
static uint8_t WriteByte (uint8_t Value)
{
   uint8_t i;
   uint8_t Error;
   
   for(i=0x80; i>0; i=i >> 1)
   {
      if (Value & i)
      {
         DATA_HIGH();
      }
      else
      {
         DATA_LOW();
      }
      WAIT_DATA();
      
      CLK_HIGH();
      WAIT();
      CLK_LOW();
      WAIT();
   }
   
   /* Release DATA-line */
   DATA_HIGH();
   WAIT_DATA();
   
   CLK_HIGH();
   WAIT();
   
   /* Check ack (DATA will be pulled down by the bmp085) */
   Error = DATA_GET();
   
   CLK_LOW();
   WAIT();

   return(Error);
} /* WriteByte */

/*************************************************************************/
/*  ReadByte                                                             */
/*                                                                       */
/*  Reads a byte form the bus and gives an acknowledge in case           */
/*  of CheckACK == 1                                                     */
/*                                                                       */
/*  Note: Tested with a STM32F103 where the code was running in flash    */
/*        and the CPU clock was 72MHz.                                   */
/*                                                                       */
/*  In    : CheckACK                                                     */
/*  Out   : none                                                         */
/*  Return: Value                                                        */
/*************************************************************************/
static uint8_t ReadByte (uint8_t CheckACK)
{
   uint8_t i;
   uint8_t Value = 0;
   
   /* Release DATA-line */
   DATA_HIGH();
   
   /* Read value */
   for(i=0x80; i>0; i = i >> 1)
   {
      CLK_HIGH();
      WAIT_DATA();
      if (DATA_GET())
      {
         Value = Value | i;
      }
      CLK_LOW();
      WAIT();
      WAIT();
      WAIT();
   }
   
   /* Check ACK must set */
   if (1 == CheckACK)
   {
      DATA_LOW();
   }
   WAIT();
   CLK_HIGH();
   WAIT();
   CLK_LOW();
   WAIT();
   
   /* Set data back to low */
   DATA_LOW();
   WAIT_DATA();
   
   return(Value);
} /* ReadByte */

/*************************************************************************/
/*  ReadU24                                                              */
/*                                                                       */
/*  Read a 24 bits from the given address.                               */
/*                                                                       */
/*  In    : Addr, pData                                                  */
/*  Out   : none                                                         */
/*  Return: 0 = no error / 1 = error                                     */
/*************************************************************************/
static uint8_t ReadU24 (uint8_t Addr, uint32_t *pData)
{
   uint8_t  Error = 0;
   uint8_t  Msb; 
   uint8_t  Lsb;
   uint8_t  Xlsb;
   uint32_t Value;

   START();
   Error |= WriteByte(BMP_ADDR_WRITE); /* Module address, write */
   Error |= WriteByte(Addr);           /* Register address */
   START();
   Error |= WriteByte(BMP_ADDR_READ);  /* Module address, write */
   
   Msb  = ReadByte(ACK);
   Lsb  = ReadByte(ACK);
   Xlsb = ReadByte(NO_ACK);
   
   STOP();
   
   Value  = Msb;
   Value  = Value << 8;
   Value |= Lsb;
   Value  = Value << 8;
   Value |= Xlsb;
   
   *pData = Value;
   
   return(Error);   
} /* ReadU24 */

/*************************************************************************/
/*  ReadU16                                                              */
/*                                                                       */
/*  Read a 16 bits from the given address.                               */
/*                                                                       */
/*  In    : Addr, pData                                                  */
/*  Out   : none                                                         */
/*  Return: 0 = no error / 1 = error                                     */
/*************************************************************************/
static uint8_t ReadU16 (uint8_t Addr, uint16_t *pData)
{
   uint8_t  Error = 0;
   uint8_t  Msb; 
   uint8_t  Lsb;
   uint16_t Value;

   START();
   Error |= WriteByte(BMP_ADDR_WRITE); /* Module address, write */
   Error |= WriteByte(Addr);           /* Register address */
   START();
   Error |= WriteByte(BMP_ADDR_READ);  /* Module address, write */
   
   Msb = ReadByte(ACK);
   Lsb = ReadByte(NO_ACK);
   
   STOP();
   
   Value  = Msb;
   Value  = (uint16_t)(Value << 8);
   Value |= Lsb;
   
   *pData = Value;
   
   return(Error);   
} /* ReadU16 */

/*************************************************************************/
/*  HWInit                                                               */
/*                                                                       */
/*  Setup the port pins like clk and data.                               */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void HWInit (void)
{
   /* Turn on GPIOB */
   RCC->AHB1ENR |= RCC_AHB1ENR_GPIOBEN;

   /* Prevent glitch on clk */
   CLK_HIGH();

   /* Set PB10 to output (CLK) */
   GPIOB->MODER &= ~GPIO_MODER_MODER10;
   GPIOB->MODER |= GPIO_MODER_MODER10_0;

   /* Prevent glitch on data */
   DATA_HIGH();   
   
   /* Set PB11 to output, Open-drain (DATA) */
   GPIOB->MODER  &= ~GPIO_MODER_MODER11;
   GPIOB->MODER  |= GPIO_MODER_MODER11_0;
   GPIOB->OTYPER |= GPIO_OTYPER_OT_11;
   
} /* HWInit */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  bmp085_Init                                                          */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void bmp085_Init (void)
{
   HWInit();
      
   /*
    * Read calibration data
    */
   ReadU16(0xAA, (uint16_t*)&AC1);
   ReadU16(0xAC, (uint16_t*)&AC2);
   ReadU16(0xAE, (uint16_t*)&AC3);
   ReadU16(0xB0, (uint16_t*)&AC4);
   ReadU16(0xB2, (uint16_t*)&AC5);
   ReadU16(0xB4, (uint16_t*)&AC6);
   
   ReadU16(0xB6, (uint16_t*)&B1);
   ReadU16(0xB8, (uint16_t*)&B2);
   
   ReadU16(0xBA, (uint16_t*)&MB);
   ReadU16(0xBC, (uint16_t*)&MC);
   ReadU16(0xBE, (uint16_t*)&MD);
       
} /* bmp085_Init */

/*************************************************************************/
/*  bmp085_Measurement                                                   */
/*                                                                       */
/*  Return Temperature and Pressure.                                     */
/*                                                                       */
/*  Note: The altitude will be ignored for the moment.                   */
/*                                                                       */
/*  In    : Altitude, pTemperature, pPressure                            */
/*  Out   : pTemperature, pPressure                                      */
/*  Return: 0 = no error / 1 = error                                     */
/*************************************************************************/
uint8_t bmp085_Measurement (uint16_t Altitude, float *pTemperature, uint16_t *pPressure)
{
   uint8_t  Error = 0;
   uint16_t Cnt;
   uint8_t  EOC;

   uint8_t  oss = 0;
   
   int32_t  UT;
   int32_t  X1;
   int32_t  X2;
   int32_t  B5;
   int32_t  T;
   
   int32_t  UP;
   int32_t  X3;
   int32_t  B3;
   uint32_t B4;
   int32_t  B6;
   uint32_t B7;
   int32_t  P;

   (void)Altitude;


   tal_LEDSet(TAL_LED_CHANNEL_3);
   

   /*
    * Read uncompensated temperature value
    */   
   START();
   Error |= WriteByte(BMP_ADDR_WRITE);  /* Module address, write */
   Error |= WriteByte(0xF4);            /* Register address */
   Error |= WriteByte(0x2e);            /* Control register */
   STOP();   

   Cnt = 0; 
   do
   {
      Cnt++;
      EOC = EOC_GET();
   } while ((0 == EOC) && (Cnt < MAX_CONV_TEMP_TIMEOUT));

   UT = 0;
   ReadU16(0xF6, (uint16_t*)&UT);

   
   /*
    * Calculate true temperature
    */
   X1 = ((UT - AC6) * AC5) >> 15;
   X2 = (MC << 11) / ( X1 + MD);
   B5 = X1 + X2;
   T  = (B5 + 8) >> 4;
   
   *pTemperature = (float)T / 10;
   

   /*
    * Read uncompensated pressure value
    */
   START();
   Error |= WriteByte(BMP_ADDR_WRITE);  /* Module address, write */
   Error |= WriteByte(0xF4);            /* Register address */
   Error |= WriteByte(0x34+(oss<<6));   /* Control register */
   STOP();   

   Cnt = 0; 
   do
   {
      Cnt++;
      EOC = EOC_GET();
   } while ((0 == EOC) && (Cnt < MAX_CONV_PRES_TIMEOUT));

   UP = 0;
   ReadU24(0xF6, (uint32_t*)&UP);
   UP = UP >> (8-oss);
   
   
   /*
    * Calculate true pressure
    */
   B6 = B5 - 4000;
   X1 = (B2 * ((B6 * B6) >> 12)) >> 11;
   X2 = (AC2 * B6) >> 11;
   X3 = X1 + X2;
   B3 = ((((int32_t)AC1 * 4 + X3) << oss) + 2) / 4;
   X1 = (AC3 * B6) >> 13;
   X2 = (B1 * ((B6 * B6) >> 12)) >> 16;
   X3 = ((X1 + X2) + 2) >> 2;
   B4 = (AC4 * (uint32_t)(X3 + 32768)) >> 15;
   B7 = ((uint32_t)UP - B3) * (50000 >> oss);
   
   if (B7 < 0x80000000)
   {
      P = (B7 * 2) / B4;
   }
   else
   {
      P = (B7 / B4) * 2;
   }
   
   X1 = (P >> 8) * (P >> 8);
   X1 = (X1 * 3038) >> 16;
   X2 = (-7357 * P) >> 16;
   P  = P + ((X1 + X2 + 3791) >> 4);
  
   *pPressure = (P + 50) / 100;


   tal_LEDClear(TAL_LED_CHANNEL_3);

   return(Error);
} /* bmp085_Measurement */

/*** EOF ***/
